const commands = @import("./commands.zig");

const coral = @import("coral");

const ext = @import("../ext.zig");

const gfx = @import("../gfx.zig");

const sokol = @import("sokol");

const spirv = @import("./spirv.zig");

const std = @import("std");

pub const Context = struct {
	window: *ext.SDL_Window,
	thread: std.Thread,
	loop: *Loop,

	pub const Submission = struct {
		width: u16,
		height: u16,
		clear_color: gfx.Color = gfx.colors.black,
		renders: []RenderChain,
	};

	pub fn close(self: *Context, handle: *gfx.Handle) void {
		const close_commands = self.loop.closes.pending();

		std.debug.assert(close_commands.stack.cap > close_commands.stack.len());
		close_commands.append(handle) catch unreachable;
	}

	pub fn create(self: *Context, desc: gfx.Desc) gfx.OpenError!*gfx.Handle {
		const resource = try coral.heap.allocator.create(Resource);

		errdefer coral.heap.allocator.destroy(resource);

		try self.loop.creations.pending().append(.{
			.resource = resource,
			.desc = desc,
		});

		const pending_destroys = self.loop.destroys.pending();

		if (pending_destroys.stack.len() == pending_destroys.stack.cap) {
			try pending_destroys.stack.grow(1);
		}

		return @ptrCast(resource);
	}

	pub fn deinit(self: *Context) void {
		self.loop.is_running.store(false, .monotonic);
		self.loop.ready.post();
		self.thread.join();
		self.loop.deinit();
		coral.heap.allocator.destroy(self.loop);
		ext.SDL_DestroyWindow(self.window);

		self.* = undefined;
	}

	pub fn init() !Context {
		const window = create: {
			const position = ext.SDL_WINDOWPOS_CENTERED;
			const flags = ext.SDL_WINDOW_OPENGL;
			const width = 640;
			const height = 480;

			break: create ext.SDL_CreateWindow("Ona", position, position, width, height, flags) orelse {
				return error.Unsupported;
			};
		};

		errdefer ext.SDL_DestroyWindow(window);

		const loop = try coral.heap.allocator.create(Loop);

		errdefer coral.heap.allocator.destroy(loop);

		loop.* = .{};

		return .{
			.loop = loop,
			.window = window,

			.thread = spawn: {
				const thread = try std.Thread.spawn(.{}, Loop.run, .{loop, window});

				thread.setName("Ona Graphics") catch {
					std.log.warn("failed to name the graphics thread", .{});
				};

				break: spawn thread;
			},
		};
	}

	pub fn submit(self: *Context, submission: Submission) void {
		self.loop.finished.wait();

		defer self.loop.ready.post();

		for (submission.renders) |*render| {
			render.swap();
		}

		self.loop.creations.swap();
		self.loop.destroys.swap();

		var last_width, var last_height = [_]c_int{0, 0};

		ext.SDL_GetWindowSize(self.window, &last_width, &last_height);

		if (submission.width != last_width or submission.height != last_height) {
			ext.SDL_SetWindowSize(self.window, submission.width, submission.height);
		}

		self.loop.clear_color = submission.clear_color;
		self.loop.renders = submission.renders;

		self.loop.ready.post();
	}
};

const Frame = struct {
	projection: Projection = .{@splat(0), @splat(0), @splat(0), @splat(0)},
	flushed_batch_count: usize = 0,
	pushed_batch_count: usize = 0,
	render_passes: usize = 0,
	mesh: ?*gfx.Handle = null,
	texture: ?*gfx.Handle = null,
	material: ?*gfx.Handle = null,
};

const Loop = struct {
	ready: std.Thread.Semaphore = .{},
	finished: std.Thread.Semaphore = .{},
	clear_color: gfx.Color = gfx.colors.black,
	is_running: AtomicBool = AtomicBool.init(true),
	renders: []RenderChain = &.{},
	destroys: DestroyChain = DestroyChain.init(coral.heap.allocator),
	creations: CreateChain = CreateChain.init(coral.heap.allocator),
	has_resource_head: ?*Resource = null,
	has_resource_tail: ?*Resource = null,

	const AtomicBool = std.atomic.Value(bool);

	const CreateCommand = struct {
		resource: *Resource,
		desc: gfx.Desc,

		fn clone(command: CreateCommand, arena: *std.heap.ArenaAllocator) std.mem.Allocator.Error!CreateCommand {
			const allocator = arena.allocator();

			return .{
				.desc = switch (command.desc) {
					.texture => |texture| .{
						.texture = .{
							.access = switch (texture.access) {
								.static => |static| .{
									.static = .{
										.data = try allocator.dupe(coral.io.Byte, static.data),
										.width = static.width,
									},
								},

								.render => |render| .{
									.render = .{
										.width = render.width,
										.height = render.height,
									},
								},
							},

							.format = texture.format,
						},
					},

					.mesh_2d => |mesh_2d| .{
						.mesh_2d = .{
							.vertices = try allocator.dupe(gfx.Desc.Mesh2D.Vertex, mesh_2d.vertices),
							.indices = try allocator.dupe(u16, mesh_2d.indices),
						},
					},
				},

				.resource = command.resource,
			};
		}
	};

	const CreateChain = commands.Chain(CreateCommand, CreateCommand.clone);

	const DestroyChain = commands.Chain(*Resource, null);

	fn consume_creations(self: *Loop) std.mem.Allocator.Error!void {
		const create_commands = self.creations.submitted();

		defer create_commands.clear();

		for (create_commands.stack.values) |command| {
			errdefer coral.heap.allocator.destroy(command.resource);

			command.resource.* = .{
				.payload = switch (command.desc) {
					.texture => |texture| init: {
						const pixel_format = switch (texture.format) {
							.rgba8 => sokol.gfx.PixelFormat.RGBA8,
							.bgra8 => sokol.gfx.PixelFormat.BGRA8,
						};

						switch (texture.access) {
							.static => |static| {
								break: init .{
									.texture = .{
										.image = sokol.gfx.makeImage(.{
											.width = static.width,
											.height = @intCast(static.data.len / (static.width * texture.format.byte_size())),
											.pixel_format = pixel_format,

											.data = .{
												.subimage = get: {
													var subimage = [_][16]sokol.gfx.Range{.{.{}} ** 16} ** 6;

													subimage[0][0] = sokol.gfx.asRange(static.data);

													break: get subimage;
												},
											},
										}),

										.sampler = sokol.gfx.makeSampler(.{}),
									},
								};
							},

							.render => |render| {
								const color_image = sokol.gfx.makeImage(.{
									.width = render.width,
									.height = render.height,
									.render_target = true,
									.pixel_format = pixel_format,
								});

								const depth_image = sokol.gfx.makeImage(.{
									.width = render.width,
									.height = render.height,
									.render_target = true,
									.pixel_format = .DEPTH_STENCIL,
								});

								break: init .{
									.render_target = .{
										.attachments = sokol.gfx.makeAttachments(.{
											.colors = get: {
												var attachments = [_]sokol.gfx.AttachmentDesc{.{}} ** 4;

												attachments[0] = .{
													.image = color_image,
												};

												break: get attachments;
											},

											.depth_stencil = .{
												.image = depth_image,
											},
										}),

										.sampler = sokol.gfx.makeSampler(.{}),
										.color_image = color_image,
										.depth_image = depth_image,
										.width = render.width,
										.height = render.height,
									},
								};
							},
						}
					},

					.mesh_2d => |mesh_2d| init: {
						if (mesh_2d.indices.len > std.math.maxInt(u32)) {
							return error.OutOfMemory;
						}

						break: init .{
							.mesh_2d = .{
								.index_buffer = sokol.gfx.makeBuffer(.{
									.data = sokol.gfx.asRange(mesh_2d.indices),
									.type = .INDEXBUFFER,
								}),

								.vertex_buffer = sokol.gfx.makeBuffer(.{
									.data = sokol.gfx.asRange(mesh_2d.vertices),
									.type = .VERTEXBUFFER,
								}),

								.index_count = @intCast(mesh_2d.indices.len),
							},
						};
					},
				},

				.has_prev = self.has_resource_tail,
				.has_next = null,
			};

			if (self.has_resource_tail) |resource_tail| {
				resource_tail.has_next = command.resource;
			} else {
				std.debug.assert(self.has_resource_head == null);

				self.has_resource_head = command.resource;
			}

			self.has_resource_tail = command.resource;
		}
	}

	fn consume_destroys(self: *Loop) std.mem.Allocator.Error!void {
		const destroy_commands = self.destroys.submitted();

		defer destroy_commands.clear();

		for (destroy_commands.stack.values) |resource| {
			defer coral.heap.allocator.destroy(resource);

			resource.destroy();

			if (resource.has_prev) |resource_prev| {
				resource_prev.has_next = resource.has_next;
			}

			if (resource.has_next) |resource_next| {
				resource_next.has_prev = resource.has_prev;
			}
		}
	}

	fn deinit(self: *Loop) void {
		self.destroys.deinit();
		self.creations.deinit();
	}

	fn run(self: *Loop, window: *ext.SDL_Window) !void {
		const context = configure_and_create: {
			var result = @as(c_int, 0);

			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_FLAGS, ext.SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG);
			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_PROFILE_MASK, ext.SDL_GL_CONTEXT_PROFILE_CORE);
			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MAJOR_VERSION, 3);
			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MINOR_VERSION, 3);
			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_DOUBLEBUFFER, 1);

			if (result != 0) {
				return error.Unsupported;
			}

			break: configure_and_create ext.SDL_GL_CreateContext(window);
		};

		sokol.gfx.setup(.{
			.environment = .{
				.defaults = .{
					.color_format = .RGBA8,
					.depth_format = .DEPTH_STENCIL,
					.sample_count = 1,
				},
			},

			.logger = .{
				.func = sokol.log.func,
			},
		});

		defer {
			sokol.gfx.shutdown();
			ext.SDL_GL_DeleteContext(context);
		}

		var rendering_2d = try Rendering2D.init();

		defer rendering_2d.deinit();

		self.finished.post();

		while (self.is_running.load(.monotonic)) {
			self.ready.wait();

			defer self.finished.post();

			try self.consume_creations();

			var frame = Frame{};

			rendering_2d.set_target(&frame, window, .{
				.clear_color = self.clear_color,
				.clear_depth = 0,
			});

			if (self.renders.len != 0) {
				const renders_tail_index = self.renders.len - 1;

				for (self.renders[0 .. renders_tail_index]) |*render| {
					const render_commands = render.submitted();

					defer render_commands.clear();

					for (render_commands.stack.values) |command| {
						try switch (command) {
							.instance_2d => |instance_2d| rendering_2d.batch(&frame, instance_2d),
							.target => |target| rendering_2d.set_target(&frame, window, target),
						};
					}

					rendering_2d.set_target(&frame, window, .{
						.clear_color = null,
						.clear_depth = null,
					});
				}

				const render_commands = self.renders[renders_tail_index].submitted();

				defer render_commands.clear();

				for (render_commands.stack.values) |command| {
					try switch (command) {
						.instance_2d => |instance_2d| rendering_2d.batch(&frame, instance_2d),
						.target => |target| rendering_2d.set_target(&frame, window, target),
					};
				}

				rendering_2d.flush(&frame);
			}

			sokol.gfx.endPass();
			sokol.gfx.commit();
			ext.SDL_GL_SwapWindow(window);

			try self.consume_destroys();
		}

		var has_next_resource = self.has_resource_head;

		while (has_next_resource) |resource| {
			std.log.info("destroying remaining gfx device resource {x}", .{@intFromPtr(resource)});
			resource.destroy();

			has_next_resource = resource.has_next;

			coral.heap.allocator.destroy(resource);
		}
	}
};

const Projection = [4]@Vector(4, f32);

pub const RenderChain = commands.Chain(gfx.Command, clone_command);

pub const RenderList = commands.List(gfx.Command, clone_command);

const Rendering2D = struct {
	batching_pipeline: sokol.gfx.Pipeline,
	batching_buffers: coral.stack.Sequential(sokol.gfx.Buffer),

	const Instance = extern struct {
		transform: gfx.Transform2D,
		tint: @Vector(4, u8) = @splat(std.math.maxInt(u8)),
		depth: f32 = 0,
		texture_offset: gfx.Point2D = @splat(0),
		texture_size: gfx.Point2D = @splat(1),
	};

	const buffer_indices = .{
		.mesh = 0,
		.instance = 1,
	};

	fn batch(self: *Rendering2D, frame: *Frame, instance: gfx.Command.Instance2D) std.mem.Allocator.Error!void {
		if (instance.mesh_2d != frame.mesh or instance.texture != frame.texture) {
			self.flush(frame);
		}

		frame.mesh = instance.mesh_2d;
		frame.texture = instance.texture;

		const has_filled_buffer = (frame.pushed_batch_count % instances_per_buffer) == 0;
		const pushed_buffer_count = frame.pushed_batch_count / instances_per_buffer;

		if (has_filled_buffer and pushed_buffer_count == self.batching_buffers.len()) {
			var name_buffer = [_:0]u8{0} ** 64;

			const name_view = std.fmt.bufPrint(&name_buffer, "instance 2d buffer #{}", .{self.batching_buffers.len()}) catch {
				unreachable;
			};

			const instance_buffer = sokol.gfx.makeBuffer(.{
				.label = name_view.ptr,
				.size = @sizeOf(Instance) * instances_per_buffer,
				.usage = .STREAM,
			});

			errdefer sokol.gfx.destroyBuffer(instance_buffer);

			try self.batching_buffers.push_grow(instance_buffer);
		}

		_ = sokol.gfx.appendBuffer(self.batching_buffers.get().?, sokol.gfx.asRange(&Instance{
			.transform = instance.transform,
		}));

		frame.pushed_batch_count += 1;
	}

	fn deinit(self: *Rendering2D) void {
		for (self.batching_buffers.values) |buffer| {
			sokol.gfx.destroyBuffer(buffer);
		}

		self.batching_buffers.deinit();
		sokol.gfx.destroyPipeline(self.batching_pipeline);
	}

	fn flush(self: *Rendering2D, frame: *Frame) void {
		const unflushed_count = frame.pushed_batch_count - frame.flushed_batch_count;

		if (unflushed_count == 0) {
			return;
		}

		sokol.gfx.applyPipeline(self.batching_pipeline);
		sokol.gfx.applyUniforms(.VS, 0, sokol.gfx.asRange(&frame.projection));

		const mesh_2d = resource_cast(frame.mesh.?).payload.mesh_2d;

		const texture_image, const texture_sampler = switch (resource_cast(frame.texture.?).payload) {
			.texture => |texture| .{texture.image, texture.sampler},
			.render_target => |render_target| .{render_target.color_image, render_target.sampler},
			else => unreachable,
		};

		var bindings = sokol.gfx.Bindings{
			.vertex_buffers = get: {
				var buffers = [_]sokol.gfx.Buffer{.{}} ** 8;

				buffers[buffer_indices.mesh] = mesh_2d.vertex_buffer;

				break: get buffers;
			},

			.index_buffer = mesh_2d.index_buffer,

			.fs = .{
				.images = get: {
					var images = [_]sokol.gfx.Image{.{}} ** 12;

					images[0] = texture_image;

					break: get images;
				},

				.samplers = get: {
					var samplers = [_]sokol.gfx.Sampler{.{}} ** 8;

					samplers[0] = texture_sampler;

					break: get samplers;
				},
			},

			.vs = .{
				.images = get: {
					var images = [_]sokol.gfx.Image{.{}} ** 12;

					images[0] = texture_image;

					break: get images;
				},

				.samplers = get: {
					var samplers = [_]sokol.gfx.Sampler{.{}} ** 8;

					samplers[0] = texture_sampler;

					break: get samplers;
				},
			},
		};

		while (frame.flushed_batch_count < frame.pushed_batch_count) {
			const buffer_index = frame.flushed_batch_count / instances_per_buffer;
			const buffer_offset = frame.flushed_batch_count % instances_per_buffer;
			const instances_to_flush = @min(instances_per_buffer - buffer_offset, unflushed_count);

			bindings.vertex_buffers[buffer_indices.instance] = self.batching_buffers.values[buffer_index];
			bindings.vertex_buffer_offsets[buffer_indices.instance] = @intCast(@sizeOf(Instance) * buffer_offset);

			sokol.gfx.applyBindings(bindings);
			sokol.gfx.draw(0, mesh_2d.index_count, @intCast(instances_to_flush));

			frame.flushed_batch_count += instances_to_flush;
		}
	}

	fn init() spirv.Error!Rendering2D {
		sokol.gfx.setup(.{
			.environment = .{
				.defaults = .{
					.color_format = .RGBA8,
					.depth_format = .DEPTH_STENCIL,
					.sample_count = 1,
				},
			},

			.logger = .{
				.func = sokol.log.func,
			},
		});

		var spirv_unit = try spirv.Unit.init();

		defer spirv_unit.deinit();

		const shader_spirv = spirv.embed_shader("./shaders/render_2d.spv");

		try spirv_unit.compile(shader_spirv, .vertex);

		std.log.info("{s}\n", .{spirv_unit.shader_desc.vs.source});

		try spirv_unit.compile(shader_spirv, .fragment);

		std.log.info("{s}\n", .{spirv_unit.shader_desc.fs.source});

		return .{
			.batching_pipeline = sokol.gfx.makePipeline(.{
				.label = "2D drawing pipeline",

				.layout = .{
					.attrs = get: {
						var attrs = [_]sokol.gfx.VertexAttrState{.{}} ** 16;

						attrs[0] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.mesh,
						};

						attrs[1] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.mesh,
						};

						attrs[2] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.instance,
						};

						attrs[3] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.instance,
						};

						attrs[4] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.instance,
						};

						attrs[5] = .{
							.format = .UBYTE4N,
							.buffer_index = buffer_indices.instance,
						};

						attrs[6] = .{
							.format = .FLOAT,
							.buffer_index = buffer_indices.instance,
						};

						attrs[7] = .{
							.format = .FLOAT4,
							.buffer_index = buffer_indices.instance,
						};

						break: get attrs;
					},

					.buffers = get: {
						var buffers = [_]sokol.gfx.VertexBufferLayoutState{.{}} ** 8;

						buffers[buffer_indices.instance].step_func = .PER_INSTANCE;

						break: get buffers;
					},
				},

				.shader = sokol.gfx.makeShader(spirv_unit.shader_desc),
				.index_type = .UINT16,
			}),

			.batching_buffers = .{.allocator = coral.heap.allocator},
		};
	}

	const instances_per_buffer = 512;

	fn set_target(self: *Rendering2D, frame: *Frame, window: *ext.SDL_Window, target: gfx.Command.Target) void {
		defer frame.render_passes += 1;

		if (frame.render_passes != 0) {
			self.flush(frame);
			sokol.gfx.endPass();
		}

		var pass = sokol.gfx.Pass{
			.action = .{
				.stencil = .{.load_action = .CLEAR},
			},
		};

		if (target.clear_color) |color| {
			pass.action.colors[0] = .{
				.load_action = .CLEAR,
				.clear_value = @bitCast(color),
			};
		} else {
			pass.action.colors[0] = .{.load_action = .LOAD};
		}

		if (target.clear_depth) |depth| {
			pass.action.depth = .{
				.load_action = .CLEAR,
				.clear_value = depth,
			};
		} else {
			pass.action.depth = .{.load_action = .LOAD};
		}

		if (target.texture) |render_texture| {
			const render_target = resource_cast(render_texture).payload.render_target;

			pass.attachments = render_target.attachments;

			frame.projection = orthographic_projection(-1.0, 1.0, .{
				.left = 0,
				.top = 0,
				.right = @floatFromInt(render_target.width),
				.bottom = @floatFromInt(render_target.height),
			});
		} else {
			var target_width, var target_height = [_]c_int{0, 0};

			ext.SDL_GL_GetDrawableSize(window, &target_width, &target_height);
			std.debug.assert(target_width > 0 and target_height > 0);

			pass.swapchain = .{
				.width = target_width,
				.height = target_height,
				.sample_count = 1,
				.color_format = .RGBA8,
				.depth_format = .DEPTH_STENCIL,
				.gl = .{.framebuffer = 0},
			};

			frame.projection = orthographic_projection(-1.0, 1.0, .{
				.left = 0,
				.top = 0,
				.right = @floatFromInt(target_width),
				.bottom = @floatFromInt(target_height),
			});
		}

		sokol.gfx.beginPass(pass);
	}
};

const Resource = struct {
	has_prev: ?*Resource,
	has_next: ?*Resource,

	payload: union (enum) {
		mesh_2d: struct {
			index_count: u32,
			vertex_buffer: sokol.gfx.Buffer,
			index_buffer: sokol.gfx.Buffer,
		},

		texture: struct {
			image: sokol.gfx.Image,
			sampler: sokol.gfx.Sampler,
		},

		render_target: struct {
			sampler: sokol.gfx.Sampler,
			color_image: sokol.gfx.Image,
			depth_image: sokol.gfx.Image,
			attachments: sokol.gfx.Attachments,
			width: u16,
			height: u16,
		},
	},

	fn destroy(self: Resource) void {
		switch (self.payload) {
			.mesh_2d => |mesh_2d| {
				sokol.gfx.destroyBuffer(mesh_2d.vertex_buffer);
				sokol.gfx.destroyBuffer(mesh_2d.index_buffer);
			},

			.texture => |texture| {
				sokol.gfx.destroyImage(texture.image);
				sokol.gfx.destroySampler(texture.sampler);
			},

			.render_target => |render_target| {
				sokol.gfx.destroyImage(render_target.color_image);
				sokol.gfx.destroyImage(render_target.depth_image);
				sokol.gfx.destroySampler(render_target.sampler);
				sokol.gfx.destroyAttachments(render_target.attachments);
			},
		}
	}
};

fn clone_command(self: gfx.Command, arena: *std.heap.ArenaAllocator) std.mem.Allocator.Error!gfx.Command {
	_ = arena;

	return switch (self) {
		.instance_2d => |instance_2d| .{.instance_2d = instance_2d},
		.target => |target| .{.target = target},
	};
}

fn orthographic_projection(near: f32, far: f32, viewport: gfx.Rect) Projection {
	const width = viewport.right - viewport.left;
	const height = viewport.bottom - viewport.top;

	return .{
		.{2 / width, 0, 0, 0},
		.{0, 2 / height, 0, 0},
		.{0, 0, 1 / (far - near), 0},
		.{-((viewport.left + viewport.right) / width), -((viewport.top + viewport.bottom) / height), near / (near - far), 1},
	};
}

fn resource_cast(handle: *gfx.Handle) *Resource {
	return @ptrCast(@alignCast(handle));
}
